/*====================================+=====================================+
! File CDownload.cpp                  ! Copyright (C) 2016-2018 Remi PASCAL !
+-------------------------------------+-------------------------------------+
! This file is part of matelefree.                                          !
! matelefree is free software: you can redistribute it and/or modify it     !
! under the terms of the GNU General Public License as published by the     !
! Free Software Foundation, either version 3 of the License, or any later   !
! version.                                                                  !
! matelefree is distributed in the hope that it will be useful, but WITHOUT !
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or     !
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for !
! more details.                                                             !
! You should have received a copy of the GNU General Public License along   !
! with matelefree. If not, see <http://www.gnu.org/licenses/>.              !
+==========================================================================*/



/*-------------------------------------------------------------------------*/
#include <wx/thread.h>
#include <wx/wfstream.h>
#include <wx/sstream.h>
#include <wx/filename.h>
#include <curl/curl.h>
#include "CApp.h"
#include "CFrame.h"
#include "CDownload.h"
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
constexpr const char *CDownloadParam::m_st_tb_err_msg[ DE_NB ] ;
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
wxDEFINE_EVENT( EVT_DOWNLOAD, CDownloadEvent ) ;

/*-------------------------------------------------------------------------*/
CDownloadParam::~CDownloadParam()
{
   /*-----------------------------------------------------------------------+
   ! Output stream can be given as parameter and therefore is not deleted   !
   +-----------------------------------------------------------------------*/
   if( m_os != m_p_mos ) { delete m_os ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxFileName CDownloadParam::build_local_fullpath() const
{
   /*--( By default the name of the downloaded file is used )--------------*/
   if( m_fn_local.HasName() ) { return( m_fn_local ) ; }
   /*----------------------------------------------------------------------*/
   return( wxFileName( m_fn_local.GetFullPath(), m_s_url.AfterLast( '/' ) ));
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CDownloadParam::send_event( double dltotal, double dlnow ) const
{
   /*----------------------------------------------------------------------*/
   CDownloadEvent event( EVT_DOWNLOAD, wxID_ANY ) ;
   /*----------------------------------------------------------------------*/
   event.set_id( get_event_id() ) ;
   //event.set_string( "" ) ;
   event.set_err_code( get_err_code() ) ;
   event.set_current_size( dlnow ) ;
   event.set_download_size( dltotal ) ;
   /*----------------------------------------------------------------------*/
   if( wxThread::IsMain() )
   {  get_handler()->ProcessEvent( event ) ; }
   else
   {  get_handler()->QueueEvent( new CDownloadEvent( event ) ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
static int st_progress_callback( void *clientp, double dltotal, double dlnow,
                                 double ultotal, double ulnow
                               )
{
   /*----------------------------------------------------------------------*/
   CDownloadParam *p_dp = static_cast< CDownloadParam * >( clientp ) ;

   /*--( Check if stop requested )-----------------------------------------*/
   if( p_dp->get_thread() != nullptr && p_dp->get_thread()->TestDestroy() )
   {
      /*--( A partially downloaded file is closed and deleted )------------*/
      if( p_dp->get_fn_local().IsOk() )
      {  p_dp->reset_output_stream() ;
         wxRemoveFile( p_dp->get_fn_local().GetFullPath() ) ;
      }
      /*-------------------------------------------------------------------*/
      return( -1 ) ;
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   p_dp->send_event( dltotal, dlnow ) ;
   /*----------------------------------------------------------------------*/
   return( 0 ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
static size_t st_write_callback( char *ptr, size_t size, size_t nmemb,
                                 void *userdata
                               )
{
   /*----------------------------------------------------------------------*/
   CDownloadParam *p_dp = static_cast< CDownloadParam * >( userdata ) ;
   size_t sz_read = size * nmemb ;
   /*----------------------------------------------------------------------*/
   return( p_dp->get_output_stream()->Write( ptr, sz_read ).LastWrite() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CDownloadParam::e_err_code CDownloadParam::download()
{
   /*----------------------------------------------------------------------*/
   CURL *curl = curl_easy_init() ;
   /*----------------------------------------------------------------------*/
   if( curl == nullptr ) { return( DE_INIT_ERROR ) ; }
   /*----------------------------------------------------------------------*/
   wxCharBuffer cb_url = m_s_url.ToAscii() ;
   curl_easy_setopt( curl, CURLOPT_URL, cb_url.data() ) ;
   /*----------------------------------------------------------------------*/
   CDownload::st_init_proxy( curl ) ;
   /*----------------------------------------------------------------------*/
   curl_easy_setopt( curl, CURLOPT_FOLLOWLOCATION, 1 ) ;
   curl_easy_setopt( curl, CURLOPT_WRITEFUNCTION, st_write_callback ) ;
   curl_easy_setopt( curl, CURLOPT_WRITEDATA, this ) ;

   /*--( Return error on HTTP error >= 400 )-------------------------------*/
   curl_easy_setopt( curl, CURLOPT_FAILONERROR, 1 ) ;
   /*----------------------------------------------------------------------*/
   if( get_send_progress() )
   {  /*-------------------------------------------------------------------*/
      curl_easy_setopt( curl, CURLOPT_NOPROGRESS, 0 ) ;
      curl_easy_setopt( curl, CURLOPT_PROGRESSFUNCTION,
                        st_progress_callback
                      ) ;
      curl_easy_setopt( curl, CURLOPT_PROGRESSDATA, this ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   if( m_i_connect_timeout > 0 )
   {  curl_easy_setopt( curl, CURLOPT_CONNECTTIMEOUT, m_i_connect_timeout );}
   /*----------------------------------------------------------------------*/
   wxFileName fn_local ;
   /*--( Save data in a local file )---------------------------------------*/
   if( m_fn_local.IsOk() )
   {  /*-------------------------------------------------------------------*/
      m_fn_local = build_local_fullpath() ;
      m_fn_local.Mkdir( wxS_DIR_DEFAULT, wxPATH_MKDIR_FULL ) ;
      /*-------------------------------------------------------------------*/
      m_os = new wxFileOutputStream( m_fn_local.GetFullPath() ) ;
      if( m_os == nullptr || !m_os->IsOk() )
      {  m_err_code = DE_FILE_CREATION ; goto function_end ; }
      /*-------------------------------------------------------------------*/
   }
   else /*--( Save data in an output stream )------------------------------*/
   if( m_p_mos != nullptr )
   {  m_os = m_p_mos ; }
   else /*--( Save data in a string (convUTF8) )---------------------------*/
   if( m_p_s_data != nullptr )
   {  /*--( By default the string stream appends text to the string )------*/
      m_p_s_data->clear() ;
      m_os = new wxStringOutputStream( m_p_s_data ) ;
      /*-------------------------------------------------------------------*/
   }
   else /*--( By default data is appended in a string )--------------------*/
   {  m_s_def_data.clear() ;
      m_os = new wxStringOutputStream( &m_s_def_data ) ;
   }

   /*--( GO ! GO ! GO ! )--------------------------------------------------*/
   switch( curl_easy_perform( curl ) )
   {
      /*-------------------------------------------------------------------*/
      case CURLE_OK :
         m_err_code = DE_NO_ERROR_DONE ; break ;
      /*-------------------------------------------------------------------*/
      case CURLE_UNSUPPORTED_PROTOCOL :
      case CURLE_URL_MALFORMAT : m_err_code = DE_INVALID_URL ; break ;
      /*-------------------------------------------------------------------*/
      case CURLE_HTTP_RETURNED_ERROR :
      {  /*----------------------------------------------------------------*/
         long l_resp_code ;
         curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &l_resp_code ) ;
           m_err_code
         = ( l_resp_code == 404 ? DE_FILE_NOT_FOUND : DE_OTHER ) ;
         break ;
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
      case CURLE_COULDNT_RESOLVE_HOST :
         m_err_code = DE_HOST_NOT_FOUND ; break ;
      case CURLE_REMOTE_FILE_NOT_FOUND :
         m_err_code = DE_FILE_NOT_FOUND ; break ;
      case CURLE_ABORTED_BY_CALLBACK :
         m_err_code = DE_STOP_REQUEST ; break ;
      case CURLE_FAILED_INIT :
         m_err_code = DE_INIT_ERROR ; break ;
      case CURLE_COULDNT_CONNECT :
         m_err_code = DE_CONNECT ; break ;
      case CURLE_WRITE_ERROR :
         m_err_code = DE_WRITE_ERROR ; break ;
      case CURLE_READ_ERROR :
         m_err_code = DE_READ_ERROR ; break ;
      /*-------------------------------------------------------------------*/
      default : m_err_code = DE_OTHER ; break ;
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   if( m_err_code == DE_NO_ERROR_DONE )
   {
      /*--( Get last relocation )------------------------------------------*/
      char *p_c_last_url ;
      /*-------------------------------------------------------------------*/
      if( curl_easy_getinfo( curl, CURLINFO_EFFECTIVE_URL, &p_c_last_url
                           ) == CURLE_OK
        )
      {  m_s_last_url = p_c_last_url  ; }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
function_end :
   /*----------------------------------------------------------------------*/
   curl_easy_cleanup( curl ) ;
   /*----------------------------------------------------------------------*/
   return( m_err_code ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CDownload::~CDownload()
{
   /*--( The thread has to be informed that this object no longer exists )-*/
   if( m_p_thread != nullptr )
   {  m_p_thread->reset_p_dl() ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CDownload::st_init_proxy( void *curl )
{
   /*----------------------------------------------------------------------*/
   if( curl == nullptr ) { return ; }
   /*----------------------------------------------------------------------*/
   if( !wxGetApp().M_s_proxy_server.empty() )
   {
      /*-------------------------------------------------------------------*/
      wxCharBuffer cb_server = wxGetApp().M_s_proxy_server.get().ToAscii() ;
      /*-------------------------------------------------------------------*/
      curl_easy_setopt( curl, CURLOPT_PROXY, cb_server.data() ) ;
      /*--------------------------------------------------------------------+
      ! This option solves the problem with "https" in a socks proxy env :  !
      ! schannel: next InitializeSecurityContext failed:                    !
      ! Unknown error (0x80092013)                                          !
      +--------------------------------------------------------------------*/
      #ifdef __WXMSW__
         curl_easy_setopt( curl, CURLOPT_SSL_OPTIONS, CURLSSLOPT_NO_REVOKE );
      #endif
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*--( This function is called from outside of this module )----------------*/
CDownloadParam::e_err_code CDownload::non_threaded_exec()
{
   /*----------------------------------------------------------------------*/
   return( m_param.download() ) ;
   /*----------------------------------------------------------------------*/
}

/*--( This function is called from outside of this module )----------------*/
void CDownload::thread_start()
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_p_thread ) ;
   /*--( Clean current before new one )------------------------------------*/
   if( m_p_thread != nullptr ) { m_p_thread->reset_p_dl() ; }
   /*----------------------------------------------------------------------*/
   m_p_thread = new CDownloadThread( this ) ;
   /*----------------------------------------------------------------------*/
   if( m_p_thread->Create() != wxTHREAD_NO_ERROR )
   {  wxFAIL ; return ; }
   /*----------------------------------------------------------------------*/
   if( m_p_thread->Run() != wxTHREAD_NO_ERROR )
   {  wxFAIL ; return ; }
   /*----------------------------------------------------------------------*/
}

/*--( This function is called from outside of this module )----------------*/
void CDownload::thread_stop()
{
   /*----------------------------------------------------------------------*/
   m_boo_thread_stopping = true ;
   /*----------------------------------------------------------------------*/
   {  wxCriticalSectionLocker enter_csec( m_csec_p_thread ) ;
      /*-------------------------------------------------------------------*/
      if( m_p_thread != nullptr && m_p_thread->Delete() != wxTHREAD_NO_ERROR)
      {  wxFAIL ; return ; }
      /*-------------------------------------------------------------------*/
   }
   /*--( Wait until thread real death ! )----------------------------------*/
   while( true )
   {  /*-------------------------------------------------------------------*/
      {  wxCriticalSectionLocker enter_csec( m_csec_p_thread ) ;
         if( m_p_thread == nullptr ) { break ; }
      }
      /*-------------------------------------------------------------------*/
      wxThread::This()->Sleep( 100 ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   m_boo_thread_stopping = false ;
   /*----------------------------------------------------------------------*/
}

/*--( This function is called from outside of this module )----------------*/
bool CDownload::thread_running()
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_p_thread ) ;
   /*----------------------------------------------------------------------*/
   return( m_p_thread != nullptr && !m_boo_thread_stopping ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CDownloadThread::CDownloadThread( CDownload *p_dl )
                : wxThread( wxTHREAD_DETACHED ),
                  m_p_dl( p_dl ), m_param( p_dl->m_param )
{  ; }

/*-------------------------------------------------------------------------*/
CDownloadThread::~CDownloadThread()
{
   /*----------------------------------------------------------------------*/
   if( m_p_dl != nullptr )
   {  wxCriticalSectionLocker enter_csec( m_p_dl->m_csec_p_thread ) ;
      m_p_dl->m_p_thread = nullptr ;
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxThread::ExitCode CDownloadThread::Entry()
{
   /*----------------------------------------------------------------------*/
   m_param.set_p_thread( this ) ;
   m_param.download() ;
   /*--( Last event )------------------------------------------------------*/
   m_param.send_event() ;
   /*----------------------------------------------------------------------*/
   return( ( wxThread::ExitCode )0 ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CDlgDownload::CDlgDownload( wxWindow *parent, const wxString &s_url,
                            const wxFileName &fn_local, e_option ddo
                          )
            : CDlgProgress( parent ),
              m_dl( this, s_url, fn_local ),
              m_ddo( ddo )
{
   /*----------------------------------------------------------------------*/
   SetTitle( _( "Download" ) + " : " + s_url.AfterLast( '/' ) ) ;
   /*----------------------------------------------------------------------*/
   mtf::init_tlw_rect( this, wxGetApp().M_progress_rec ) ;
   Bind( EVT_DOWNLOAD, &CDlgDownload::OnEvent, this ) ;
   Layout() ;
   /*----------------------------------------------------------------------*/
   m_dl.get_param().set_send_progress( true ) ;
   m_dl.thread_start() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CDlgDownload::~CDlgDownload()
{
   /*----------------------------------------------------------------------*/
   mtf::save_tlw_rect( this, wxGetApp().M_progress_rec ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CDlgDownload::OnBtCancel( wxCommandEvent &event )
{
   /*----------------------------------------------------------------------*/
   if( m_dl.thread_running() )
   {  m_dl.thread_stop() ; }
   else
   {  EndModal( CDownloadParam::DE_STOP_REQUEST ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CDlgDownload::OnEvent( CDownloadEvent &event )
{
   /*----------------------------------------------------------------------*/
   wxString s_msg( event.get_string() ) ;
   /*----------------------------------------------------------------------*/
   if( event.get_err_code() == CDownloadParam::DE_NO_ERROR_INFO )
   {
      /*--( Message has a higher priority than pure info )-----------------*/
      if(    s_msg.empty()
          && event.get_download_size() != m_ga_progress->GetValue()
        )
      {  /*----------------------------------------------------------------*/
         wxULongLong ull_cur = event.get_current_size() ;
         wxULongLong ull_tot = event.get_download_size() ;
         s_msg =         wxFileName::GetHumanReadableSize( ull_cur )
                 + '/' + wxFileName::GetHumanReadableSize( ull_tot ) ;
         /*--( Range set before value else the visual result is not great )*/
         m_ga_progress->SetRange( ull_tot.GetLo() ) ;
         m_ga_progress->SetValue( ull_cur.GetLo() ) ;
         /*----------------------------------------------------------------*/
      }
      /*--( Limit flickering )---------------------------------------------*/
      if( s_msg != m_s_info_prev )
      {  msg_info( s_msg ) ; m_s_info_prev = s_msg ; }
      /*-------------------------------------------------------------------*/
   }
   else
   if( !s_msg.empty() )
   {  msg_error( s_msg ) ; }
   else
   if( event.get_err_code() == CDownloadParam::DE_NO_ERROR_DONE )
   {  /*--( Wait or not ? )------------------------------------------------*/
      if( m_ddo == DDO_END_WAIT )
      {  msg_info( _( "Download done" ) ) ;
         m_bt_quit->SetLabel( wxGetStockLabel( wxID_CLOSE ) ) ;
         m_bsz_gen->Layout() ;
      }
      else
      {  EndModal( event.get_err_code() ) ; }
      /*-------------------------------------------------------------------*/
   }
   else
   if( event.get_err_code() == CDownloadParam::DE_STOP_REQUEST )
   {  EndModal( event.get_err_code() ) ; }
   else
   {  msg_error( CDownloadParam::get_err_msg( event.get_err_code() ) ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CDlgDownload::msg_info( const wxString &s )
{
   /*----------------------------------------------------------------------*/
   m_stx_info->SetForegroundColour( *wxBLUE ) ;
   m_stx_info->SetLabel( s ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CDlgDownload::msg_error( const wxString &s )
{
   /*----------------------------------------------------------------------*/
   m_stx_info->SetForegroundColour( *wxRED ) ;
   m_stx_info->SetLabel( s ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CDlgDownload::go()
{
   /*----------------------------------------------------------------------*/
   return( ShowModal() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/



/*==========================================================================+
!                       End of file CDownload.cpp                           !
+==========================================================================*/
