/*====================================+=====================================+
! File CVlcMedia.cpp                  ! Copyright (C) 2016-2018 Remi PASCAL !
+-------------------------------------+-------------------------------------+
! This file is part of matelefree.                                          !
! matelefree is free software: you can redistribute it and/or modify it     !
! under the terms of the GNU General Public License as published by the     !
! Free Software Foundation, either version 3 of the License, or any later   !
! version.                                                                  !
! matelefree is distributed in the hope that it will be useful, but WITHOUT !
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or     !
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for !
! more details.                                                             !
! You should have received a copy of the GNU General Public License along   !
! with matelefree. If not, see <http://www.gnu.org/licenses/>.              !
+==========================================================================*/



/*-------------------------------------------------------------------------*/
#include <wx/wx.h>
#ifdef __WXMSW__
   #include <wx/msw/registry.h>
#elif defined( __WXGTK__ )
   #include <gdk/gdkx.h>
   #include <gtk/gtk.h>
#else
   #error Unmanaged platform to: include libvlc files
#endif
/*-------------------------------------------------------------------------*/
#include <wx/imaglist.h>
#include <wx/wupdlock.h>
/*-------------------------------------------------------------------------*/
#include "CApp.h"
#include "CFrame.h"
#include "CVlcMedia.h"
#include "CDlgPrefs.h"
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
wxDEFINE_EVENT( EVT_VLC_PLAYER, wxCommandEvent ) ;
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
static const libvlc_event_type_t st_co_tb_mp_event[]
= { libvlc_MediaPlayerOpening,
    libvlc_MediaPlayerPlaying,
    libvlc_MediaPlayerPaused,
    libvlc_MediaPlayerEncounteredError,
    libvlc_MediaPlayerEndReached
  } ;
/*-------------------------------------------------------------------------*/



/*--------------------------------------------------------------------------+
! This window size is set to zero to avoid it to be seen if the             !
! preferences are "used"                                                    !
+--------------------------------------------------------------------------*/
CVlcMedia::CVlcMedia( wxWindow *parent, const wxString &s_channel )
         : wxWindow( parent, wxID_ANY, wxDefaultPosition, wxSize( 0, 0 ),
                     wxFULL_REPAINT_ON_RESIZE | wxWANTS_CHARS
                   )
{
   /*----------------------------------------------------------------------*/
   init_channel( s_channel ) ;
   create_vlc_interface() ;
   if( m_vlc == nullptr ) { return ; }
   /*----------------------------------------------------------------------*/
   m_panel_box = new wxPanel( parent ) ;
   m_panel_box->SetName( "panel_box" ) ;
   SetForegroundColour( *wxYELLOW ) ;
   SetBackgroundColour( *wxBLACK ) ;
   /*----------------------------------------------------------------------*/
   m_bsz_box = new wxBoxSizer( wxVERTICAL ) ;
   m_panel_box->SetSizer( m_bsz_box ) ;
   Reparent( m_panel_box ) ;
   SetName( "vlcmedia" ) ;
   /*----------------------------------------------------------------------*/
   SetFont( GetParent()->GetFont().Scaled( 2 ).Bold() ) ;
   /*----------------------------------------------------------------------*/
   m_bsz_gen = new wxBoxSizer( wxVERTICAL ) ;
   SetSizer( m_bsz_gen ) ;
   m_bsz_info = new wxBoxSizer( wxHORIZONTAL ) ;
   wxBoxSizer *bsz_msg = new wxBoxSizer( wxVERTICAL ) ;

   /*-----------------------------------------------------------------------+
   ! The logo is displayed in a "frame" for a better looking display        !
   ! especially if the image had an alpha channel                           !
   +-----------------------------------------------------------------------*/
   CChannel *p_channel = get_channel_played() ;
   if( p_channel != nullptr )
   {
      /*--------------------------------------------------------------------+
      ! A frame is displayed around the image to have a nicer look for logo !
      ! having a square shape :                                             !
      ! - left and right bars with fixed colour                             !
      ! - transparent left and right bars replaced by "load_logo" function  !
      ! Some methods don't seem to work correctly on MSW or GTK :           !
      ! - wxStaticBitmap with a different size than the logo (GTK)          !
      ! - wxGenericBitmap with a different background color (MSW)           !
      +--------------------------------------------------------------------*/
      const int co_i_frame = 5 ;
      wxBitmap bmp_framed(   p_channel->get_logo().GetSize()
                           + wxSize( 2 * co_i_frame, 2 * co_i_frame )
                         ) ;
      /*-------------------------------------------------------------------*/
      wxMemoryDC dc( bmp_framed ) ;
      dc.SetBackground( wxSystemSettings::GetColour( wxSYS_COLOUR_WINDOW ) );
      dc.Clear() ;
      dc.DrawBitmap( p_channel->get_logo(), co_i_frame, co_i_frame ) ;
      dc.SelectObject( wxNullBitmap ) ;
      /*-------------------------------------------------------------------*/
      m_stb_logo = new wxStaticBitmap( this, wxID_ANY, bmp_framed ) ;
      m_bsz_info->Add( m_stb_logo, 0, wxALL | wxALIGN_CENTER_VERTICAL, 5 ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   m_stx_msg = new wxStaticText( this, wxID_ANY, wxEmptyString ) ;
   bsz_msg->Add( m_stx_msg, 0, wxALL, 5 ) ;
   /*--( Colors have to be forced to replace system ones )-----------------*/
     m_ga_retry_timer
   = new wxGauge( this, wxID_ANY, 0, wxDefaultPosition, wxSize( -1, 12 ) ) ;
   m_ga_retry_timer->SetForegroundColour( GetForegroundColour() ) ;
   m_ga_retry_timer->SetBackgroundColour( GetBackgroundColour() ) ;
   bsz_msg->Add( m_ga_retry_timer, 0, wxALL | wxEXPAND, 5 ) ;
   m_ga_retry_timer->Hide() ;
   /*----------------------------------------------------------------------*/
   m_bsz_info->Add( bsz_msg, 0, wxALIGN_CENTER_VERTICAL ) ;
   /*--( By default, nothing is displayed )--------------------------------*/
   m_bsz_info->ShowItems( false ) ;
   /*----------------------------------------------------------------------*/
   m_bsz_gen->AddStretchSpacer() ;
   m_bsz_gen->Add( m_bsz_info, 0, wxALIGN_CENTER_HORIZONTAL ) ;
   m_bsz_gen->AddStretchSpacer() ;
   /*--( Frame to show status )--------------------------------------------*/
   m_bsz_box->Add( this, 1, wxEXPAND | wxALL, 7 ) ;
   m_panel_box->Layout() ;
   show_focus() ;
   /*----------------------------------------------------------------------*/
   Bind( EVT_VLC_PLAYER, &CVlcMedia::manage_media_event, this ) ;
   Bind( wxEVT_KEY_DOWN, &CVlcMedia::OnKeyDown, this ) ;
   /*----------------------------------------------------------------------*/
   Bind( wxEVT_SIZE,
         [ this ]( wxSizeEvent &event ) { show_message() ; }
       ) ;
   /*-----------------------------------------------------------------------+
   ! wxWidgets propagates only wxCommandEvent events to the parents         !
   ! Therefore they are "bound" manually to all the family                  !
   +-----------------------------------------------------------------------*/
   bind_mouse_event( wxEVT_LEFT_DOWN, &CVlcMedia::OnMouseLeftDown ) ;
   bind_mouse_event( wxEVT_LEFT_DCLICK, &CVlcMedia::OnMouseLeftDClick ) ;
   bind_mouse_event( wxEVT_MIDDLE_DOWN, &CVlcMedia::OnMouseMiddleDown ) ;
   bind_mouse_event( wxEVT_RIGHT_DOWN, &CVlcMedia::OnMouseRightDown ) ;
   /*----------------------------------------------------------------------*/
   Bind( wxEVT_SET_FOCUS,
         [ this ]( wxFocusEvent &event ) { show_focus() ; }
       ) ;
   Bind( wxEVT_KILL_FOCUS,
         [ this ]( wxFocusEvent &event ) { show_focus() ; }
       ) ;
   /*-----------------------------------------------------------------------+
   ! The mute/unmute, no video for radio etc. is not easy to set and to be  !
   ! taken into account by vlc.                                             !
   ! These timers will try to set all this correctly. Not nice, I know !    !
   +-----------------------------------------------------------------------*/
   m_timer_init_audio.SetOwner( this ) ;
   Bind( wxEVT_TIMER,
         [ this ]( wxTimerEvent &event ) { init_audio() ; },
         m_timer_init_audio.GetId()
       ) ;
   m_timer_init_video.SetOwner( this ) ;
   Bind( wxEVT_TIMER,
         [ this ]( wxTimerEvent &event ) { init_video() ; },
         m_timer_init_video.GetId()
       ) ;
   /*-----------------------------------------------------------------------+
   ! Some channels, like the twitch ones, may be offline when the user      !
   ! wants to play them. When asked to be played this timer will check      !
   ! regularly if it has become online and then starts playing it           !
   +-----------------------------------------------------------------------*/
   m_timer_retry_play.SetOwner( this ) ;
   Bind( wxEVT_TIMER, &CVlcMedia::OnPlayRetry,
         this, m_timer_retry_play.GetId()
       ) ;
   /*----------------------------------------------------------------------*/
   m_timer_message.SetOwner( this ) ;
   Bind( wxEVT_TIMER,
         [ this ]( wxTimerEvent &event ) { show_message() ; },
         m_timer_message.GetId()
       ) ;
   /*----------------------------------------------------------------------*/
   m_mp = m_vlc->m_libvlc_media_player_new( m_vlc->get_inst() ) ;
   all_evt_mp_attach() ;

   /*--( We want to handle the user events directly (not let them to vlc) )*/
   set_video_mouse_input( false ) ;
   set_video_key_input( false ) ;
   /*----------------------------------------------------------------------*/
   set_deinterlace() ;
   /*--( For an explanation of this test see function CFrame::OnFirstIdle )*/
   if( wxGetApp().m_frame->get_first_idle_done() )
   {  init_player_window() ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CVlcMedia::~CVlcMedia()
{
   /*--( Delete the interface and the player )-----------------------------*/
   delete_vlc_interface() ;
   /*--( For a nicer display the last message is removed )-----------------*/
   clear_message() ;
   /*----------------------------------------------------------------------*/
   if( m_panel_box != nullptr )
   {  Reparent( m_panel_box->GetParent() ) ;
      delete m_panel_box ;
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::OnKeyDown( wxKeyEvent &event )
{
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   /*----------------------------------------------------------------------*/
   switch( event.GetKeyCode() )
   {
      /*-------------------------------------------------------------------*/
      case WXK_SPACE  : frame->play_start_pause() ; return ;
      case 'F'        : frame->next_fullscreen() ; return ;
      case WXK_ESCAPE : frame->set_fullscreen( FSL_NONE ) ; return ;
      case WXK_DELETE : frame->close_media( this, true ) ; return ;
      /*-------------------------------------------------------------------*/
      case WXK_UP :
      case WXK_DOWN :
      case WXK_LEFT :
      case WXK_RIGHT :
      case WXK_HOME :
      case WXK_END :
         change_media_active( event.GetKeyCode() ) ;
         return ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   event.Skip() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::OnMouseLeftDown( wxMouseEvent &event )
{
   /*----------------------------------------------------------------------*/
   wxGetApp().m_frame->set_media_active( this ) ;
   SetFocus() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::OnMouseLeftDClick( wxMouseEvent &event )
{
   /*----------------------------------------------------------------------*/
   wxGetApp().m_frame->next_fullscreen() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::OnMouseMiddleDown( wxMouseEvent &event )
{
   /*----------------------------------------------------------------------*/
   wxGetApp().m_frame->close_media( this, true ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::OnMouseRightDown( wxMouseEvent &event )
{
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   /*--( The right clicked media becomes the active one )------------------*/
   frame->set_media_active( this ) ;
   SetFocus() ;
   /*----------------------------------------------------------------------*/
   CBoolInScope bool_scope( m_boo_popup_menu_on ) ;
   /*--( A menubar menu can't remain attached while being used elsewhere )-*/
   wxMenu menu ;
   mtf::append_menu( &menu, frame->get_menu_play() ) ;
   frame->PopupMenu( &menu ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::OnPlayRetry( wxTimerEvent &event )
{
   /*----------------------------------------------------------------------*/
   m_ga_retry_timer->SetValue( ++m_i_wait_sec_retry ) ;
   /*----------------------------------------------------------------------*/
   if( m_i_wait_sec_retry >= m_ga_retry_timer->GetRange() )
   {  set_message_nowait( _( "Retrying ..." ) ) ;
      play_channel() ;
   }
   else
   {  m_timer_retry_play.StartOnce( 1000 ) ;
      Update() ;
   }
   /*----------------------------------------------------------------------*/
}

/*--( Mouse events are not propagated to the parent : we "force" them )----*/
void CVlcMedia::bind_mouse_event( t_mouse_event_type eventType,
                                  t_mouse_event_func functor
                                )
{
   /*--( Bind the event to this window )-----------------------------------*/
   Bind( eventType, functor, this ) ;
   /*--( And to all its children )-----------------------------------------*/
   for( auto win : GetChildren() )
   {  win->Bind( eventType, functor, this ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::create_vlc_interface()
{
   /*--( The vlc init string depends on the channel )----------------------*/
   CChannel *p_channel = get_channel_played() ;
   wxString s_vlc_init = wxGetApp().build_vlc_init( p_channel ) ;

   /*--( Try to dynamically link to VLC )----------------------------------*/
   m_vlc = new CVlcInterface( wxGetApp().M_s_vlc_path.get(), s_vlc_init ) ;

   /*--( If not possible then force the user to give the right path )------*/
   if( !m_vlc->is_ok() )
   {
      /*--( Loaded but another problem occurred like wrong version )-------*/
      if( m_vlc->IsLoaded() )
      {  delete_vlc_interface() ; }
      /*--( Ask the user to change the vlc path )--------------------------*/
      if( CDlgPrefs().get_modal() != wxID_OK )
      {  delete_vlc_interface() ;
         return ;
      }
      /*--( Normally here all should be ok ... )---------------------------*/
      m_vlc = new CVlcInterface( wxGetApp().M_s_vlc_path.get(), s_vlc_init );
      if( !m_vlc->is_ok() )
      {  delete_vlc_interface() ;
         return ;
      }
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   m_vlc->set_user_agent( MTF_DISPLAY_NAME_ASCII " " MTF_VERSION,
                          MTF_DISPLAY_NAME_ASCII "/VLC/wxWidgets"
                         ) ;
   m_vlc->set_app_id( "net.scarabee-software." MTF_NAME,
                      MTF_VERSION, MTF_NAME
                    ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::delete_vlc_interface()
{
   /*----------------------------------------------------------------------*/
   if( m_vlc == nullptr ) { return ; }
   /*----------------------------------------------------------------------*/
   if( m_mp != nullptr )
   {  stop_channel() ;
      m_vlc->m_libvlc_media_player_release( m_mp ) ;
      m_mp = nullptr ;
   }
   /*----------------------------------------------------------------------*/
   delete m_vlc ;
   m_vlc = nullptr ;
   /*----------------------------------------------------------------------*/
}

/*--( Send event interface between VLC and our object )--------------------*/
static void st_manage_media_event( const struct libvlc_event_t *p_event,
                                   void *p_data
                                 )
{
   /*-----------------------------------------------------------------------+
   ! The function is called from another thread than the main one.          !
   ! As some display is done in it an event is sent to the main thread.     !
   +-----------------------------------------------------------------------*/
   wxQueueEvent( static_cast< CVlcMedia * >( p_data ),
                 new wxCommandEvent( EVT_VLC_PLAYER, p_event->type )
               ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
bool CVlcMedia::all_evt_mp_attach()
{
   /*----------------------------------------------------------------------*/
   if( m_vlc == nullptr ) { return( false ) ; }
   /*----------------------------------------------------------------------*/
   libvlc_event_manager_t *evt_man = get_media_player_evt_man() ;
   bool boo_ok = true ;
   /*----------------------------------------------------------------------*/
   for( const auto evt_type : st_co_tb_mp_event )
   {
       /*------------------------------------------------------------------*/
       if( m_vlc->m_libvlc_event_attach( evt_man, evt_type,
                                         st_manage_media_event, this
                                       ) != 0
         )
       {  boo_ok = false ; }
       /*------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return( boo_ok ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::init_player_window()
{
   /*----------------------------------------------------------------------*/
   if( m_vlc == nullptr ) { return ; }
   /*-----------------------------------------------------------------------+
   ! Using this windows to display media                                    !
   +-----------------------------------------------------------------------*/
   #ifdef __WXMSW__
      m_vlc->m_libvlc_media_player_set_hwnd( m_mp, GetHandle() ) ;
   #elif defined( __WXGTK__ )
      /*-------------------------------------------------------------------*/
      GtkWidget *gtk_widget = GetHandle() ;
      gtk_widget_realize( gtk_widget ) ;
      Window window_id = GDK_WINDOW_XWINDOW( gtk_widget->window ) ;
      m_vlc->m_libvlc_media_player_set_xwindow( m_mp, window_id ) ;
      /*-------------------------------------------------------------------*/
   #else
      #error Unmanaged platform to: initialize libvlc media player window
   #endif
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::reset_player_window()
{
   /*----------------------------------------------------------------------*/
   if( m_vlc == nullptr ) { return ; }
   /*----------------------------------------------------------------------*/
   #ifdef __WXMSW__
      m_vlc->m_libvlc_media_player_set_hwnd( m_mp, nullptr ) ;
   #elif defined( __WXGTK__ )
      m_vlc->m_libvlc_media_player_set_xwindow( m_mp, 0 ) ;
   #else
      #error Unmanaged platform to: reset libvlc media player window
   #endif
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::init_channel( const wxString &s_channel )
{
   /*----------------------------------------------------------------------*/
   if( s_channel.empty() ) { return ; }
   /*----------------------------------------------------------------------*/
   m_s_channel = s_channel ;
   /*----------------------------------------------------------------------*/
   CPlaylist *pl = wxGetApp().m_frame->get_playlist_active() ;
   /*--( Current playlist contains played channel ? )----------------------*/
   if( pl != nullptr && pl->get_channel_item( s_channel ) >= 0 )
   {  /*--------------------------------------------------------------------+
      ! The playlist name is kept to be able to find back the currently     !
      ! channel played. The name is stored rather than a pointer because    !
      ! the playlist could have been deleted ...                            !
      +--------------------------------------------------------------------*/
      m_s_pl_name = pl->get_name() ;
      /*-------------------------------------------------------------------*/
   }
   else
   {  /*--( The channel "source" playlist becomes the "played" one )-------*/
      CChannel *p_channel = get_channel_played() ;
      if( p_channel != nullptr ) { m_s_pl_name = p_channel->get_pl_source();}
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::refresh_pl_channel()
{
   /*----------------------------------------------------------------------*/
   if( m_s_channel.empty() ) { return ; }
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   CPlaylist *pl = frame->get_playlist_active() ;
   /*----------------------------------------------------------------------*/
   if( pl != nullptr ) { pl->refresh_item( m_s_channel ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::manage_media_event( wxCommandEvent &event )
{
   /*----------------------------------------------------------------------*/
   if( event.GetId() == libvlc_MediaPlayerOpening )
   {
      /*-------------------------------------------------------------------*/
      set_message( _( "Opening ..." ) ) ;
      /*-------------------------------------------------------------------*/
   }
   else
   if( event.GetId() == libvlc_MediaPlayerPlaying )
   {
      /*--------------------------------------------------------------------+
      ! No info for video ... it does it itself                             !
      ! Sometimes, during a play done at startup, the "radio" test or the   !
      ! video deactivation don't seem to work ...                           !
      +--------------------------------------------------------------------*/
      if( is_media_radio() )
      {  set_message_nowait( _( "Playing ..." ) ) ; }
      else
      {  clear_message() ; }
      /*-------------------------------------------------------------------*/
      CFrame *frame = wxGetApp().m_frame ;
      /*-------------------------------------------------------------------*/
      init_audio() ;
      init_video() ;
      /*-------------------------------------------------------------------*/
      if( this == frame->get_media_active() )
      {
         /*----------------------------------------------------------------*/
         frame->add_play_hist( get_channel_played() ) ;
         frame->init_gui_play() ;
         /*----------------------------------------------------------------*/
      }
      /*--( The focus color changes when the video is playing )------------*/
      show_focus() ;
      /*-------------------------------------------------------------------*/
   }
   else
   if( event.GetId() == libvlc_MediaPlayerPaused )
   {
      /*--------------------------------------------------------------------+
      ! Radio media don't show video : they don't have one or, like for     !
      ! the Freebox ones, we don't show it.                                 !
      ! Therefore we have to update the displayed message.                  !
      +--------------------------------------------------------------------*/
      if( is_media_radio() && event.GetId() == libvlc_MediaPlayerPaused )
      {  set_message_nowait( _( "Paused" ) ) ; }
      /*-------------------------------------------------------------------*/
   }
   else /*--( Happens when a network stream "stops" )----------------------*/
   if( event.GetId() == libvlc_MediaPlayerEndReached )
   {
      /*-------------------------------------------------------------------*/
      CFrame *frame = wxGetApp().m_frame ;
      /*-------------------------------------------------------------------*/
      if( this == frame->get_media_active() )
      {  frame->init_gui_play() ; }
      /*-------------------------------------------------------------------*/
      if( !get_play_retry() )
      {  stop() ;
         set_message_nowait( _( "Offline" ) ) ;
      }
      else
      {  /*----------------------------------------------------------------*/
          CChannel *p_channel = get_channel_played() ;
          if( p_channel != nullptr )
          {  /*--( For "live" channels the streams have to be reloaded )---*/
             p_channel->reset_streams() ;
             play_channel() ;
             /*------------------------------------------------------------*/
          }
          /*---------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }
   else
   if( event.GetId() == libvlc_MediaPlayerEncounteredError )
   {
      /*-------------------------------------------------------------------*/
      set_message_nowait( _( "Access error" ) ) ;
      /*-------------------------------------------------------------------*/
      CFrame *frame = wxGetApp().m_frame ;
      /*-------------------------------------------------------------------*/
      if( this == frame->get_media_active() )
      {  frame->init_gui_play() ; }
      /*--( For "live" channels the streams may have to be reloaded )------*/
      CChannel *p_channel = get_channel_played() ;
      if( p_channel != nullptr ) { p_channel->reset_streams() ; }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
bool CVlcMedia::is_media_radio()
{
   /*----------------------------------------------------------------------*/
   CChannel *p_channel = get_channel_played() ;
   /*--( Freebox radio have a video track ... )----------------------------*/
   return(    p_channel != nullptr
           && (    p_channel->get_pl_known() == PK_FBRADIO
                || (    get_audio_track_count() > 0
                     && get_video_track_count() == 0
                   )
              )
         ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::init_audio()
{
   /*----------------------------------------------------------------------*/
   CChannel *p_channel = get_channel_played() ;
   if( p_channel == nullptr ) { return ; }
   /*----------------------------------------------------------------------*/
   const str_stream *p_stream = p_channel->get_stream( this ) ;
   if( p_stream == nullptr ) { return ; }
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   bool boo_mute = (    wxGetApp().M_boo_mute.get()
                     || frame->get_media_active() != this
                   ) ;

   /*--( The first init has to be checked ... )----------------------------*/
   ++m_i_num_init_audio ;
   bool boo_to_check = ( m_i_num_init_audio == 1 ) ;

   /*--( Find out track number to set )------------------------------------*/
   int i_track = get_audio_track() ;
   if( boo_mute && wxGetApp().M_boo_mute_deact.get() )
   {  i_track = -1 ; }
   else
   if( p_stream->m_i_audio_track >= 0 )
   {  i_track = p_stream->m_i_audio_track ; }
   else
   if( p_channel->m_i_audio_track != CChannel::DEFAULT_TRACK )
   {  i_track = p_channel->m_i_audio_track ; }

   /*--( Change audio track if needed )------------------------------------*/
   if( get_audio_track() != i_track )
   {  set_audio_track( i_track ) ;
      boo_to_check = true ;
   }
   /*--( Mute / Unmute ? )-------------------------------------------------*/
   if( get_mute() != boo_mute )
   {  set_mute( boo_mute ) ;
      boo_to_check = true ;
   }
   /*----------------------------------------------------------------------*/
   if( !boo_mute )
   {  set_volume( wxGetApp().M_i_volume.get() ) ; }

   /*--( Rerunning if not correctly set or media not parsed yet )----------*/
   if(    !m_timer_init_video.IsRunning()
       && ( boo_to_check || get_all_track_count() == 0 )
     )
   {  /*-------------------------------------------------------------------*/
      m_timer_init_audio.StartOnce( wxGetApp().M_i_check_media_delay.get() );
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::init_video()
{
   /*----------------------------------------------------------------------*/
   CChannel *p_channel = get_channel_played() ;
   if( p_channel == nullptr ) { return ; }

   /*--( The first init has to be checked ... )----------------------------*/
   ++m_i_num_init_video ;
   bool boo_to_check = ( m_i_num_init_video == 1 ) ;

   /*--( Find out track number to set )------------------------------------*/
   int i_track = get_video_track() ;
   if( is_media_radio() )
   {  i_track = -1 ; }
   else
   if( p_channel->m_i_video_track != CChannel::DEFAULT_TRACK )
   {  i_track = p_channel->m_i_video_track ; }

   /*--( Change video track if needed )------------------------------------*/
   if( get_video_track() != i_track )
   {  set_video_track( i_track ) ;
      boo_to_check = true ;
   }

   /*----------------------------------------------------------------------*/
   if(    p_channel->m_i_video_spu != CChannel::DEFAULT_TRACK
       && get_video_spu() != p_channel->m_i_video_spu
     )
   {  set_video_spu( p_channel->m_i_video_spu ) ;
      boo_to_check = true ;
   }

   /*--( Rerunning if not correctly set or media not parsed yet )----------*/
   if(    !m_timer_init_video.IsRunning()
       && ( boo_to_check || get_all_track_count() == 0 )
     )
   {  /*-------------------------------------------------------------------*/
      m_timer_init_video.StartOnce( wxGetApp().M_i_check_media_delay.get() );
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::set_media_by_location( const wxString &s_url )
{
   /*----------------------------------------------------------------------*/
   if( m_vlc == nullptr ) { return ; }
   /*----------------------------------------------------------------------*/
   libvlc_media_t *media
   = m_vlc->m_libvlc_media_new_location( m_vlc->get_inst(), s_url.mb_str() );
   /*----------------------------------------------------------------------*/
   m_vlc->m_libvlc_media_player_set_media( m_mp, media ) ;
   m_vlc->m_libvlc_media_release( media ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::play_channel()
{
   /*----------------------------------------------------------------------*/
   CChannel *p_channel = get_channel_played() ;
   if( p_channel == nullptr ) { return ; }
   /*--( This is a new play )----------------------------------------------*/
   m_i_num_init_audio = 0 ;
   m_i_num_init_video = 0 ;

   /*--( Ok to played right now ? )----------------------------------------*/
   const str_stream *p_stream = p_channel->get_stream( this ) ;
   if( p_stream != nullptr && p_channel->is_gs_no_error() )
   {
      /*-------------------------------------------------------------------*/
      m_boo_play_retry = false ;
      m_ga_retry_timer->Hide() ;
      /*--( A stream is available )----------------------------------------*/
      set_media_by_location( p_stream->m_s_url ) ;
      /*--( The "m3u" "#EXTVLCOPT" are added before play )-----------------*/
      for( const wxString &s_option : p_stream->m_vs_option )
      {  add_vlc_option( s_option ) ; }
      /*-------------------------------------------------------------------*/
      play() ;
      /*-------------------------------------------------------------------*/
   }
   else /*--( Try again later ? )------------------------------------------*/
   if( p_channel->is_gs_retry_later() )
   {
      /*--( During the first retry the message is displayed )--------------*/
      if( !m_boo_play_retry )
      {  /*----------------------------------------------------------------*/
         m_boo_play_retry = true ;
         wxGetApp().m_frame->init_gui_play() ;
         /*----------------------------------------------------------------*/
      }
      /*--( The focus color changes when the video is "running" )----------*/
      show_focus() ;
      /*-------------------------------------------------------------------*/
      m_i_wait_sec_retry = 0 ;
      m_ga_retry_timer->Show() ;
      m_ga_retry_timer->SetRange( wxGetApp().M_i_retry_delay.get() ) ;
      m_ga_retry_timer->SetValue( m_i_wait_sec_retry ) ;
      m_timer_retry_play.StartOnce( 1000 ) ;
      set_message_nowait( _( "Offline" ) ) ;
      /*-------------------------------------------------------------------*/
   }
   else /*--( Display current error message )------------------------------*/
   {  set_message_nowait( p_channel->get_gs_error_message() ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::stop_channel()
{
   /*----------------------------------------------------------------------*/
   set_message( _( "Stopping ..." ) ) ;
   stop() ;
   /*----------------------------------------------------------------------*/
}

/*--( Show playlist and item of channel currently played in this media )---*/
void CVlcMedia::show_played_channel_item()
{
   /*--( Enough info to continue ? )---------------------------------------*/
   if( m_s_channel.empty() || m_s_pl_name.empty() ) { return ; }
   /*----------------------------------------------------------------------*/
   CBookPlaylist *bopl = wxGetApp().m_frame->m_bopl ;
   /*--( Search playlist from which the channel has been started (if any) )*/
   wxTreeItemId ti = bopl->find_ti_name( m_s_pl_name ) ;
   if( !ti.IsOk() ) { return ; }
   /*----------------------------------------------------------------------*/
   CPlaylist *pl = bopl->get_playlist( ti ) ;
   if( pl == nullptr ) { return ; }
   /*--( We've the playlist, now search the channel in it )----------------*/
   long l_item = pl->get_channel_item( m_s_channel ) ;
   if( l_item < 0 ) { return ; }
   /*--( Let's show all )--------------------------------------------------*/
   bopl->set_selection( ti ) ;
   pl->set_item_active( l_item ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::clear_message()
{
   /*----------------------------------------------------------------------*/
   m_timer_message.Stop() ;
   m_s_message.clear() ;
   /*----------------------------------------------------------------------*/
   if( m_stx_msg != nullptr && m_stx_msg->IsShown() )
   {  m_bsz_info->ShowItems( false ) ;
      Update() ;
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::set_message_nowait( const wxString &s_msg )
{
   /*----------------------------------------------------------------------*/
   m_timer_message.Stop() ;
   m_s_message = s_msg ;
   /*----------------------------------------------------------------------*/
   show_message() ;
   /*----------------------------------------------------------------------*/
}

/*--------------------------------------------------------------------------+
! This function is used to display a message after a certain amount of time !
! to avoid useless flickering of texts if information has to be displayed   !
! and the task between them is very short                                   !
+--------------------------------------------------------------------------*/
void CVlcMedia::set_message( const wxString &s_msg )
{
   /*----------------------------------------------------------------------*/
   m_timer_message.Stop() ;
   m_s_message = s_msg ;
   /*----------------------------------------------------------------------*/
   m_timer_message.StartOnce( 300 ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::show_message()
{
   /*----------------------------------------------------------------------*/
   if( m_stx_msg == nullptr || m_s_message.empty() ) { return ; }
   /*----------------------------------------------------------------------*/
   CChannel *p_channel = get_channel_played() ;
   /*--( Display changes are useless to be showed )------------------------*/
   wxWindowUpdateLocker no_window_updates( this ) ;
   /*----------------------------------------------------------------------*/
   if( m_stb_logo != nullptr && p_channel != nullptr )
   {  m_stb_logo->Show() ; }
   m_stx_msg->SetFont( GetFont() ) ;
   m_stx_msg->SetLabel( p_channel == nullptr
                        ? m_s_message
                        : p_channel->get_title() + " : " + m_s_message
                      ) ;
   m_stx_msg->Show() ;
   Layout() ;
   /*--( Message has to fit available space and remain readable )----------*/
   while(    m_bsz_info->GetSize().GetWidth() >= GetSize().GetWidth()
          && m_stx_msg->GetFont().GetPointSize() >= 6
        )
   {  m_stx_msg->SetFont( m_stx_msg->GetFont().Smaller() ) ;
      Layout() ;
   }
   /*----------------------------------------------------------------------*/
   Update() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::show_focus()
{
   /*--( Only one media ? => a border is useless )-------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   wxSizerItem *szi = m_bsz_box->GetItem( static_cast< size_t >( 0 ) ) ;
   if( frame->nb_media_to_show() == 1 )
   {  /*-------------------------------------------------------------------*/
      szi->SetFlag( szi->GetFlag() & ~wxALL ) ;
      m_panel_box->Layout() ;
      return ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   szi->SetFlag( szi->GetFlag() | wxALL ) ;
   m_panel_box->Layout() ;
   /*----------------------------------------------------------------------*/
   wxColor color ;
   /*----------------------------------------------------------------------*/
   if( this == frame->get_media_active() ) // Dark Yellow 2
   {  color = ( HasFocus() ? *wxYELLOW : wxColor( 153, 153, 0 ) ) ; }
   else
   if( m_mp == nullptr || !is_running() ) // Dark Yellow 4
   {  color = wxColor( 51, 51, 0 ) ; }
   else
   {  color = *wxBLACK ; }
   /*----------------------------------------------------------------------*/
   m_panel_box->SetBackgroundColour( color ) ;
   m_panel_box->ClearBackground() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CVlcMedia::change_media_active( int i_key_code )
{
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   int i_current = frame->find_media( this ) ;
   if( i_current < 0 ) { return ; }
   /*----------------------------------------------------------------------*/
   wxSizer *sizer_media = frame->get_sizer_media() ;
   int i_nb_row = sizer_media->GetItemCount() ;
   if( i_nb_row == 0 ) { return ; }
   /*----------------------------------------------------------------------*/
   int i_nb_col = sizer_media->GetItem( static_cast< size_t >( 0 )
                                      )->GetSizer()->GetItemCount() ;
   if( i_nb_col == 0 ) { return ; }
   /*----------------------------------------------------------------------*/
   int i_last = static_cast< int >( frame->get_vec_media().size() ) - 1 ;
   int i_new  = i_current ;
   /*----------------------------------------------------------------------*/
   switch( i_key_code )
   {
      /*-------------------------------------------------------------------*/
      case WXK_UP :
         /*--( On first line : go to the last one on the same column )-----*/
         if( i_new >= i_nb_col )
         {  i_new -= i_nb_col ; }
         else
         {  i_new += i_nb_col * ( i_nb_row - 1 ) ; }
         break ;
      /*-------------------------------------------------------------------*/
      case WXK_DOWN :
         /*--( On last line : go to the first one on the same column )-----*/
         if( i_new >= i_nb_col * ( i_nb_row - 1 ) )
         {  i_new %= i_nb_col ; }
         else
         {  i_new += i_nb_col ; }
         break ;
      /*-------------------------------------------------------------------*/
      case WXK_LEFT :
         /*--( On first column : go to the last one on the same line )-----*/
         if( i_new % i_nb_col == 0 )
         {  i_new += i_nb_col - 1 ; }
         else
         {  --i_new ; }
         break ;
      /*-------------------------------------------------------------------*/
      case WXK_RIGHT :
         /*--( On last column : go to the first one on the same line )-----*/
         if( i_new == i_last || i_new % i_nb_col == i_nb_col - 1 )
         {  i_new = i_new / i_nb_col * i_nb_col ; }
         else
         {  ++i_new ; }
         break ;
      /*-------------------------------------------------------------------*/
      case WXK_HOME :
         /*--( Go to first column of the same line )-----------------------*/
         i_new = i_new / i_nb_col * i_nb_col ;
         break ;
      /*-------------------------------------------------------------------*/
      case WXK_END :
         /*--( Goto to last column of the same line )----------------------*/
         i_new = i_new / i_nb_col * i_nb_col + i_nb_col - 1 ;
         break ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   i_new = wxMin( wxMax( i_new, 0 ), i_last ) ;
   /*----------------------------------------------------------------------*/
   if( i_current != i_new )
   {  frame->set_media_active( i_new ) ;
      frame->get_media_active()->SetFocus() ;
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/



/*==========================================================================+
!                         End of file CVlcMedia.cpp                         !
+==========================================================================*/
